import math
from .vector2d import Vector2d
from abc import abstractmethod
from abc import abstractclassmethod

class Block():
    def __init__(self, vector_x: float, vector_y: float, type_id: int, is_solid: bool=False, is_moveable: bool=False):
        self._vector2d = Vector2d(vector_x, vector_y)
        self.__type_id = type_id
        self._is_solid = is_solid
        self.__is_moveable = is_moveable
        self._set_x_y(vector_x, vector_y)

    @property
    def x(self):
        return self._vector2d.x

    @property
    def y(self):
        return self._vector2d.y

    @property
    def block_x(self):
        return self.__block_x

    @property
    def block_y(self):
        return self.__block_y

    @property
    def offset_x(self):
        return self.__offset_x

    @property
    def offset_y(self):
        return self.__offset_y

    @property
    def type_id(self):
        return self.__type_id

    @property
    def is_solid(self):
        return self._is_solid

    @property
    def is_moveable(self):
        return self.__is_moveable

    def _set_x_y(self, vector_x: float, vector_y: float):
        self._vector2d = Vector2d(vector_x, vector_y)
        self.__block_x = math.floor(vector_x)
        self.__block_y = math.floor(vector_y)
        self.__offset_x = vector_x % 1
        self.__offset_y = vector_y % 1

    @abstractmethod
    def _get_bounds(self, fov_ang: float):
        pass

    def is_in_fov(self, pos: Vector2d, fov_ang: float, fov_delta: float, dist: float):
        bounds = self._get_bounds(fov_ang)
        # fov_ang_min = (fov_ang - (fov_delta * 0.5) + math.pi * 2) % (math.pi * 2)
        # fov_ang_max = (fov_ang + (fov_delta * 0.5) + math.pi * 2) % (math.pi * 2)
        # fov_ang_min, fov_ang_max = (fov_ang_min, fov_ang_max) if fov_ang_min < fov_ang_max else (fov_ang_max, fov_ang_min)
        for bound in bounds:
            bound_pos = (bound - pos) ** -fov_ang
            if dist > bound_pos.mag:
                bound_ang = bound_pos.ang - math.pi * 2 if bound_pos.ang > math.pi else bound_pos.ang
                if fov_delta * -0.5 <= bound_ang <= fov_delta * 0.5:
                    return True
        return False

    
    def interacted(self):
        pass

    def get_state(self, player):
        return 0



import math
from .block import Block
from .vector2d import Vector2d
from .config import DOOR_OPEN_VELOCITY
from .config import DOOR_INTERVAL_CLOSE
from .animation import Animation

class Door(Block):
    def __init__(self, x: float, y: float, type_id: int, is_vertical: bool):
        super().__init__(x, y, type_id, True)
        self.__is_vertical = is_vertical
        self._is_solid = True
        self.__is_opened = False
        self.__is_opening = False
        self.__is_closing = False
        self.__original_x = x
        self.__original_y = y
        self.__closing_interval = 0
        self.__animations = {
            "open": Animation(DOOR_OPEN_VELOCITY, self.__on_opening_animate, self.__on_opening_animate_end),
            "wait": Animation(DOOR_INTERVAL_CLOSE, on_animate_end=self.__on_waiting_animate_end),
            "close": Animation(DOOR_OPEN_VELOCITY, self.__on_closing_animate, self.__on_closing_animate_end)
        }


    @property
    def is_vertical(self):
        return self.__is_vertical

    def __open(self):
        self.__animations["open"].start()

    def __close(self):
        self._is_solid = True
        self.__animations["close"].start()

    def __on_opening_animate(self, factor):
        vector = Vector2d(factor if not self.is_vertical else 0, factor if self.is_vertical else 0)
        self._vector2d = Vector2d(self.__original_x, self.__original_y) - vector

    def __on_opening_animate_end(self):
        self.__is_opening = False
        self.__is_opened = True
        self._is_solid = False
        self.__animations["wait"].start()

    def __on_waiting_animate_end(self):
        self.__close()

    def __on_closing_animate(self, factor):
        vector = Vector2d(1 - factor if not self.is_vertical else 0, 1 - factor if self.is_vertical else 0)
        self._vector2d = Vector2d(self.__original_x, self.__original_y) - vector

    def __on_closing_animate_end(self):
        self.__is_closing = False
        self.__is_opened = False


    def _get_bounds(self, fov_ang: float):
        bounds = []
        bounds.append(self._vector2d)
        bounds.append(self._vector2d + Vector2d(1 if not self.__is_vertical else 0, 1 if self.__is_vertical else 0))
        return bounds

    def update(self, delta_time: float, persons: []):
        for anim in self.__animations.values():
            anim.update(delta_time)
                
    def interacted(self):
        if not self.__is_opening and not self.__is_closing:
            if not self.__is_opened:
                self.__open()



import math
from .vector2d import Vector2d
from .config import RAY_COUNT
from .config import FOV_ANGLE
from .ray import Ray


class FieldOfView():
    def __init__(self, angle: float):
        self.__ang_abs_min = (FOV_ANGLE * -0.5 + math.pi * 2) % (math.pi * 2)
        self.__ang_abs_max = (FOV_ANGLE * 0.5 + math.pi * 2) % (math.pi * 2)
        self.__vector2d_ang_min = Vector2d(1, 0) ** self.__ang_abs_min
        self.__vector2d_ang_max = Vector2d(1, 0) ** self.__ang_abs_max
        self.__vector2d_ang = Vector2d(1, 0)
        self.__dist = 0
        ray_count = RAY_COUNT - RAY_COUNT % 2
        ray_angle = FOV_ANGLE / (RAY_COUNT - 1)
        self.__rays = [Ray((c * ray_angle) - (FOV_ANGLE / 2)) for c in range(ray_count)]

        self.rot(angle)

    @property
    def ang(self):
        return self.__vector2d_ang.ang

    @property
    def ang_min(self):
        return self.__vector2d_ang_min.ang

    @property
    def ang_max(self):
        return self.__vector2d_ang_max.ang

    @property
    def rays(self):
        return tuple(self.__rays)

    def rot(self, rad: float):
        self.__vector2d_ang = self.__vector2d_ang ** rad
        self.__vector2d_ang_min = self.__vector2d_ang_min ** rad
        self.__vector2d_ang_max = self.__vector2d_ang_max ** rad
        for ray in self.__rays:
            ray.rot(rad)

    def cast(self, player, grid):
        pos = Vector2d(player.x, player.y)
        for ray in self.__rays:
            ray.cast_wall(pos, grid)
            if ray.dist > self.__dist:
                self.__dist = ray.dist

        dist = max([ray.dist for ray in self.__rays])
        doors = grid.get_doors_by_fov(pos, self.ang, FOV_ANGLE, dist)
        for ray in self.__rays:
            ray.cast_doors(pos, doors)
            if ray.dist > self.__dist:
                self.__dist = ray.dist

        dist = max([ray.dist for ray in self.__rays])
        items = grid.get_items_by_fov(pos, self.ang, FOV_ANGLE, dist)
        for ray in self.__rays:
            ray.cast_items(player, items)



import math
from .block import Block
from .vector2d import Vector2d
from .field_of_view import FieldOfView
from .config import PERSON_TURN_VELOCITY
from .config import PERSON_MOVEMENT_VELOCITY
from .weapon import WeaponPistol

class Person(Block):
    def __init__(self, vector_x: float, vector_y: float, type_id: int, fov: FieldOfView):
        super().__init__(vector_x, vector_y, type_id, True, True)
        self.__fov = fov
        self.__last_pos = self._vector2d.copy()
        self.is_turning_left = False
        self.is_turning_right = False
        self.is_moving_front = False
        self.is_moving_back = False
        self.is_moving_left = False
        self.is_moving_right = False
        self.is_interacting = False
        self._weapon = WeaponPistol()

    @property
    def fov(self):
        return self.__fov

    @property
    def fov_ang(self):
        return self.__fov.ang

    @property
    def weapon(self):
        return self._weapon;

    def update(self, delta_time: float, grid):
        if self.is_turning_left:
            self.turn(-PERSON_TURN_VELOCITY * delta_time)
        if self.is_turning_right:
            self.turn(PERSON_TURN_VELOCITY * delta_time)
        movement_x = 0
        movement_y = 0
        movement_count = 0
        if self.is_moving_front:
            movement_x += PERSON_MOVEMENT_VELOCITY * delta_time
            movement_count += 1
        if self.is_moving_back:
            movement_x -= PERSON_MOVEMENT_VELOCITY * delta_time
            movement_count += 1
        if self.is_moving_right:
            movement_y += PERSON_MOVEMENT_VELOCITY * delta_time
            movement_count += 1
        if self.is_moving_left:
            movement_y -= PERSON_MOVEMENT_VELOCITY * delta_time
            movement_count += 1
        if movement_x != 0 or movement_y != 0:
            self.move(Vector2d(movement_x / movement_count, movement_y / movement_count))
        if self.is_interacting:
            self.interact(grid)
        self._weapon.update(delta_time, self, grid)

    def can_shoot(self):
        return True

    def shoot(self):
        pass

    def adjust_collision(self, grid):
        diff_pos = self._vector2d - self.__last_pos
        item = grid.get_block(math.floor(self._vector2d.x + math.copysign(0.5, diff_pos.x)), math.floor(self.__last_pos.y))
        if item is not None and item.is_solid:
            if diff_pos.x > 0:
                self._vector2d = Vector2d(item.block_x - 0.5, self._vector2d.y)
            elif diff_pos.x < 0:
                self._vector2d = Vector2d(item.block_x + 1.5, self._vector2d.y)

        item = grid.get_block(math.floor(self.__last_pos.x), math.floor(self._vector2d.y + math.copysign(0.5, diff_pos.y)))
        if item is not None and item.is_solid:
            if diff_pos.y > 0:
                self._vector2d = Vector2d(self._vector2d.x, item.block_y - 0.5)
            elif diff_pos.y < 0:
                self._vector2d = Vector2d(self._vector2d.x, item.block_y + 1.5)

    def cast(self, grid):
        self.__fov.cast(self, grid)

    def move(self, movement: Vector2d):
        self.__last_pos = self._vector2d.copy()
        self._vector2d = ((self._vector2d ** -self.__fov.ang) + movement) ** self.__fov.ang

    def turn(self, angle: float):
        self.__fov.rot(angle)

    def interact(self, grid):
        view_dir = (Vector2d.create_with_ang(self.fov.ang) % 0.75) + self._vector2d
        block = grid.get_block(view_dir.x, view_dir.y)
        if block is not None:
            block.interacted()



import math
import numpy as np
from .item import Item
from .vector2d import Vector2d
from .decoration import Decoration
from .wall import Wall
from .door import Door
from .block import Block
from .enemy import Enemy


class ItemGrid():
    def __init__(self, blocks: []):
        self.__max_x = max(map(lambda i: i.block_x, blocks)) + 1
        self.__max_y = max(map(lambda i: i.block_y, blocks)) + 1
        grid = [None] * self.__max_y * self.__max_x
        self.__items = []
        self.__doors = []
        self.__enemies = []
        self.__grid = np.asarray([grid[n:n+self.__max_y]
                                  for n in range(0, len(grid), self.__max_y)])
        for block in blocks:
            if isinstance(block, Item):
                self.__items.append(block)
            if isinstance(block, Door):
                self.__doors.append(block)
            if isinstance(block, Enemy):
                self.__enemies.append(block)
            if not block.is_moveable:
                self.__grid[block.block_x, block.block_y] = block

    def get_block(self, block_x, block_y) -> Item:
        block_x = math.floor(block_x)
        block_y = math.floor(block_y)
        if block_x < self.max_x and block_y < self.max_y:
            return self.__grid[block_x, block_y]
        return None

    def remove_block(self, block_x, block_y):
        block_x = math.floor(block_x)
        block_y = math.floor(block_y)
        self.__grid[block_x, block_y] = None

    def get_blocks_by_fov(self, pos: Vector2d, fov_ang: float, fov_delta: float, dist: float, type_block = Block):
        blocks = self.__grid
        fov_ang_vec_min = Vector2d.create_with_ang(fov_ang - (fov_delta * 0.5))
        fov_ang_vec_max = Vector2d.create_with_ang(fov_ang + (fov_delta * 0.5))
        if fov_ang_vec_min.x > 0 and fov_ang_vec_max.x > 0:
            blocks = blocks[math.floor(pos.x):, :]
        elif fov_ang_vec_min.x < 0 and fov_ang_vec_max.x < 0:
            blocks = blocks[:math.floor(pos.x+1), :]
        if fov_ang_vec_min.y > 0 and fov_ang_vec_max.y > 0:
            blocks = blocks[:, math.floor(pos.y):]
        elif fov_ang_vec_min.y < 0 and fov_ang_vec_max.y < 0:
            blocks = blocks[:, :math.floor(pos.y+1)]

        blocks = [block for line in blocks for block in line
                      if isinstance(block, type_block)]

        blocks = [block for block in blocks if block.is_in_fov(pos, fov_ang, fov_delta, dist)]

        return blocks

    def get_doors_by_fov(self, pos: Vector2d, fov_ang: float, fov_delta: float, dist: float):
        return self.get_blocks_by_fov(pos, fov_ang, fov_delta, dist, Door)

    def get_items_by_fov(self, pos: Vector2d, fov_ang: float, fov_delta: float, dist: float):
        blocks = self.get_blocks_by_fov(pos, fov_ang, fov_delta, dist, Item)
        blocks.extend(self.get_enemies_by_fov(pos, fov_ang, fov_delta, dist))
        return blocks

    def get_enemies_by_fov(self, pos: Vector2d, fov_ang: float, fov_delta: float, dist: float):
        blocks = [block for block in self.__enemies if block.is_in_fov(pos, fov_ang, fov_delta, dist)]
        return blocks

    @property
    def max_x(self):
        return self.__max_x

    @property
    def max_y(self):
        return self.__max_y



"""contem classe Vector2d"""

import math

class Vector2d():
    """permite realizar calculos vetoriais"""
    def __init__(self, x: float = 0, y: float = 0):
        self.__x = x
        self.__y = y
        self.__ang = (math.atan2(self.__y, self.__x) + math.pi * 2) % (math.pi * 2)
        self.__mag = math.sqrt(math.pow(self.__x, 2) + math.pow(self.__y, 2))

    def get_x(self) -> float:
        """obtem x"""
        return self.__x
    x: float = property(get_x)

    def get_y(self) -> float:
        """obtem y"""
        return self.__y
    y: float = property(get_y)

    def get_mag(self) -> float:
        """obtem magnitude"""
        return self.__mag
    mag: float = property(get_mag)

    def get_ang(self) -> float:
        """obtem angulo"""
        return self.__ang
    ang = property(get_ang)

    def copy(self):
        """cria copia"""
        return Vector2d(self.x, self.y)

    def __add__(self, other):
        return Vector2d(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        return Vector2d(self.x - other.x, self.y - other.y)

    def __mul__(self, other: float):
        return Vector2d(self.x * other, self.y * other)

    def __truediv__(self, other: float):
        return Vector2d(self.x / other, self.y / other)

    def __floordiv__(self, other: float):
        return Vector2d(self.x // other, self.y // other)

    def __pow__(self, other: float):
        cos = math.cos(other)
        sin = math.sin(other)
        v_x = self.x
        v_y = self.y
        return Vector2d(v_x * cos - v_y * sin, v_x * sin + v_y * cos)

    def __mod__(self, other: float):
        mag = self.mag
        if mag == 0:
            return self
        return Vector2d((self.__x / mag) * other, (self.__y / mag) * other)

    @staticmethod
    def create_with_ang(rad: float):
        return Vector2d(math.cos(rad), math.sin(rad))



from .item import Item
from .player import Player
from .weapon import WeaponSubmachine
from .weapon import WeaponMinigun

class ItemWeapon(Item):
    def touch(self, player: Player, grid):
        if self.type_id == 85:
            player.change_weapon(WeaponSubmachine)
        elif self.type_id == 86:
            player.change_weapon(WeaponMinigun)
        grid.remove_block(self.x, self.y)


import time
import asyncio
from threading import Thread
from .map2d import Map2d
from .config import FRAME_PER_SECONDS

class Server:
    def __init__(self):
        self.__map2d: Map2d
        self.__game_is_running = False
        self.__thread: Thread = None
        self.frame_count = 0
        self.__delta_time = 1 / FRAME_PER_SECONDS

    def load_map_file(self, file_path:str):
        pattern: str
        with open(file_path, "r") as file:
            pattern = "".join(file.readlines())
        self.__map2d = Map2d.create_with_pattern(pattern)

    def start_game(self):
        if not self.__game_is_running:
            # if self.__map2d is None:
            #     raise ValueError("Map2d")
            self.__game_is_running = True
            self.__thread = Thread(target=self.__loop)
            self.__thread.setDaemon(True)
            self.__thread.start()

    # def __start_loop(self):
    #     asyncio.set_event_loop(self.__loop)
    #     self.__loop.run_until_complete(self.__clock())
    #     self.__thread.

    def stop_game(self):
        self.__game_is_running = False
        if self.__thread is not None:
            self.__thread.join()

    def get_player_state(self, player_id: str):
        return self.__map2d.get_player(player_id)

    def player_start_moving_front(self, player_id: str):
        self.__map2d.get_player(player_id).is_moving_front = True
    def player_start_moving_back(self, player_id: str):
        self.__map2d.get_player(player_id).is_moving_back = True
    def player_start_moving_right(self, player_id: str):
        self.__map2d.get_player(player_id).is_moving_right = True
    def player_start_moving_left(self, player_id: str):
        self.__map2d.get_player(player_id).is_moving_left = True
    def player_start_turning_right(self, player_id: str):
        self.__map2d.get_player(player_id).is_turning_right = True
    def player_start_turning_left(self, player_id: str):
        self.__map2d.get_player(player_id).is_turning_left = True
    def player_start_interacting(self, player_id: str):
        self.__map2d.get_player(player_id).is_interacting = True
    def player_start_shooting(self, player_id: str):
        self.__map2d.get_player(player_id).weapon.is_shooting = True

    def player_stop_moving_front(self, player_id: str):
        self.__map2d.get_player(player_id).is_moving_front = False
    def player_stop_moving_back(self, player_id: str):
        self.__map2d.get_player(player_id).is_moving_back = False
    def player_stop_moving_right(self, player_id: str):
        self.__map2d.get_player(player_id).is_moving_right = False
    def player_stop_moving_left(self, player_id: str):
        self.__map2d.get_player(player_id).is_moving_left = False
    def player_stop_turning_right(self, player_id: str):
        self.__map2d.get_player(player_id).is_turning_right = False
    def player_stop_turning_left(self, player_id: str):
        self.__map2d.get_player(player_id).is_turning_left = False
    def player_stop_interacting(self, player_id: str):
        self.__map2d.get_player(player_id).is_interacting = False
    def player_stop_shooting(self, player_id: str):
        self.__map2d.get_player(player_id).weapon.is_shooting = False

    def update(self, delta_time):
        self.__delta_time = delta_time
        self.__map2d.update(self.__delta_time)

    def __loop(self):
        self.frame_count += 1
        loop_time = 1 / FRAME_PER_SECONDS
        while self.__game_is_running:
            start_time = time.time()
            self.update(loop_time)
            # print(start_time)
            end_time = time.time()
            self.__delta_time = end_time - start_time
            # time.sleep(1)
            if self.__delta_time < loop_time:
                time.sleep(loop_time - self.__delta_time)
            else:
                time.sleep(0.00000001)



import math
import os
import numpy as np
from PIL import Image as pil


class Image():
    def __init__(self):
        self.__images = [None] * 255
        self.__add_wall()
        self.__add_items()
        self.__add_weapons()
        self.__add_enemy_guard()
        self.__add_enemy_dog()

    def __add_wall(self):
        img = pil.open(f'{os.path.dirname(__file__)}/../imgs/walls.png').convert('RGBA')
        img_size = 64
        for i in range(55):
            img_x = i % 3 * 2 * img_size
            img_y = math.floor(i / 3) * img_size
            self.__images[i] = []
            self.__images[i].append(pil.Image.crop(img, (img_x, img_y, img_x + img_size, img_y + img_size)));
            self.__images[i].append(pil.Image.crop(img, (img_x + img_size, img_y, img_x + img_size * 2 ,img_y + img_size)));

    def __add_items(self):
        img = pil.open(f'{os.path.dirname(__file__)}/../imgs/items.png').convert('RGBA')
        img_size = 64
        for i in range(64):
            img_x = (i % 5 * img_size + i % 5)
            img_y = math.floor(i / 5) * img_size + math.floor(i / 5)
            self.__images[56 + i] = []
            self.__images[56 + i].append(pil.Image.crop(img, (img_x, img_y, img_x + img_size, img_y + img_size)));

    def __add_weapons(self):
        img = pil.open(f'{os.path.dirname(__file__)}/../imgs/weapons.png').convert('RGBA')
        img_size = 64
        for i in range(4):
            img_x = 0
            img_y = i * img_size
            self.__images[120 + i] = []
            for j in range(5):
                x1 = j + img_size * j
                x2 = j + img_size * (j + 1)
                self.__images[120 + i].append(pil.Image.crop(img, (x1, img_y, x2, img_y + img_size)));

    def __add_enemy_guard(self):
        img = pil.open(f'{os.path.dirname(__file__)}/../imgs/enemy-guard.png').convert('RGBA')
        img_size = 64
        self.__images[130] = []
        for i in range(7):
            img_y = img_size * i + i
            for j in range(8):
                img_x = img_size * j + j
                self.__images[130].append(pil.Image.crop(img, (img_x, img_y, img_x + img_size, img_y + img_size)));

    def __add_enemy_dog(self):
        img = pil.open(f'{os.path.dirname(__file__)}/../imgs/enemy-dog.png').convert('RGBA')
        img_size = 64
        self.__images[131] = []
        for i in range(7):
            img_y = img_size * i + i
            for j in range(8):
                img_x = img_size * j + j
                self.__images[131].append(pil.Image.crop(img, (img_x, img_y, img_x + img_size, img_y + img_size)));


    def get(self, type_id: int, width=64, height=64, state=0):
        img = self.__images[type_id][state]
        if width != 64 or height != 64:
            img = pil.Image.resize(img, (width, height))
        return img

    def get_column(self, type_id: int, col: int, height=64.0, state=0):
        img = self.__images[type_id][state]
        img = pil.Image.crop(img, (col, 0, col + 1, 64))
        img = pil.Image.resize(img, (1, round(height)), pil.NEAREST)
        return img

    @staticmethod
    def create_background(width, height):
        img = pil.new("RGBA", (width, height))
        pixels = [(30, 30, 30, 255)] * math.floor(width * height / 2)
        pixels += [(100, 100, 100, 255)] * math.ceil(width * height / 2)
        img.putdata(pixels)
        return img



import os
import math
import numpy as np
from PIL import Image as pil_image
from ..engine import FieldOfView
from ..engine import Ray
from ..engine import Player
from ..engine.config import RAY_COUNT
from .image import Image

class Screen:
    def __init__(self, images: Image):
        self.__images = images
        self.__hud_img = pil_image.open(
            f'{os.path.dirname(__file__)}/../imgs/hud.png')
        self.__hud_img.load()
        self.__hud_img = np.asarray(self.__hud_img, dtype="int32")
        self.__screen_w = RAY_COUNT
        self.__screen_h = math.floor(RAY_COUNT * 0.5)
        self.__hud_h = math.floor(RAY_COUNT / 8)
        self.__image_size = 64
        self.__pixel_template = "\033[48;2;{};{};{}m  "
        self.__image = None


    def draw(self, console, player: Player):
        pixel_matrix = self.create_pixel_matrix(player)
        console.WriteConsole("\033[0;0H")
        console_text = []
        for pixel_h in range(self.__screen_h):
            console_line = []
            for pixel_w in range(self.__screen_w):
                pixel = pixel_matrix[pixel_h][pixel_w]
                console_pixel = self.__pixel_template.format(pixel[0], pixel[1], pixel[2])
                console_line.append(console_pixel)
            console_text.append("".join(console_line))
        console.WriteConsole("\n".join(console_text))

    def create_pixel_matrix(self, player: Player):
        fov = player.fov
        img = Image.create_background(self.__screen_w, self.__screen_h)
        for i in range(self.__screen_w):
            ray = fov.rays[i]
            if ray.type_id is None:
                continue
            if ray.dist_adjusted > 0:
                height = round(self.__screen_h * 2 / ray.dist_adjusted)
                offset = math.floor(ray.offset * 64)
                img_column = self.__images.get_column(ray.type_id, offset, height, 1 if ray.is_vertical else 0)
                img.paste(img_column, (i, math.floor(self.__screen_h / 2 - height / 2)))
                list_items = list(ray.items + ray.doors)
                if len(list_items) > 0:
                    list_items.sort(key=lambda i: i.dist, reverse=True)
                    for ray_item in list_items:
                        if ray_item.dist > 0.001:
                            height = round(self.__screen_h * 2 / ray_item.dist)
                            if type(ray_item).__name__ == "RayDoor":
                                offset = math.floor(ray_item.offset * 64)
                                img_column = self.__images.get_column(49, offset, height, ray_item.state)
                            else:
                                offset = math.floor((ray_item.offset + 0.5) * 64)
                                img_column = self.__images.get_column(ray_item.type_id, offset, height, ray_item.state)
                            img.paste(img_column, (i, math.floor(self.__screen_h / 2 - height / 2)), img_column)
        img_weapon = self.__images.get(player.weapon.type_id)
        if player.weapon.shoot_animation.is_animating:
            if player.weapon.shoot_animation.factor < 0.2:
                img_weapon = self.__images.get(player.weapon.type_id, state=1)
            elif player.weapon.shoot_animation.factor < 0.3:
                img_weapon = self.__images.get(player.weapon.type_id, state=2)
            elif player.weapon.shoot_animation.factor < 0.6:
                img_weapon = self.__images.get(player.weapon.type_id, state=3)
            else:
                img_weapon = self.__images.get(player.weapon.type_id, state=4)
        img.paste(img_weapon, (round(self.__screen_w / 2 - img_weapon.width / 2), self.__screen_h - img_weapon.height), img_weapon)
        return np.asarray(img)

    def draw_hud(self, console):
        console.WriteConsole(f"\033[{self.__screen_h};0H")
        pixel_matrix = self.get_image_hud()
        img_h = 40
        img_factor = img_h / (RAY_COUNT / 8)
        console_text = []
        for pixel_h in range(self.__hud_h):
            console_line = []
            for pixel_w in range(self.__screen_w):
                pixel = pixel_matrix[math.floor(pixel_h * img_factor)][math.floor(pixel_w * img_factor)]
                console_pixel = self.__pixel_template.format(pixel[0], pixel[1], pixel[2])
                console_line.append(console_pixel)
            console_text.append("".join(console_line))
        console.WriteConsole("\n".join(console_text))


    def get_image_hud(self):
        return self.__hud_img[:40, :, :]


import uuid
from .person import Person
from .field_of_view import FieldOfView
from .vector2d import Vector2d
from .item_grid import ItemGrid
from .weapon import Weapon
from .weapon import WeaponKnife
from .weapon import WeaponPistol
from .item import Item


class Player(Person):
    def __init__(self, vector_x: float, vector_y: float, type_id: int, fov: FieldOfView, player_id: str):
        super().__init__(vector_x, vector_y, type_id, fov)
        self.__player_id = player_id
        self.__health = 100
        self.__ammo = 8
        self.__score = 0
        self.change_weapon(WeaponPistol)
        

    @property
    def health(self):
        return self.__health

    @property
    def ammo(self):
        return self.__ammo

    @property
    def score(self):
        return self.__score

    @property
    def player_id(self):
        return self.__player_id

    def shoot(self):
        self.__ammo -= 1
        self.__ammo = self.__ammo if self.__ammo >= 0 else 0
        if self.__ammo <= 0 and not isinstance(self._weapon, WeaponKnife):
            self._weapon = WeaponKnife()

    def change_weapon(self, weapon_type):
        self.__weapon_type = weapon_type
        if self.__ammo > 0:
            self._weapon = self.__weapon_type()

    def add_ammo(self, quantity):
        self.__ammo += quantity
        self.__ammo = self.__ammo if self.__ammo < 999 else 999
        if self.__ammo > 0:
            self._weapon = self.__weapon_type()

    def add_score(self, quantity):
        self.__score += quantity

    def update(self, delta_time: float, grid: ItemGrid):
        super().update(delta_time, grid)
        block = grid.get_block(self.x, self.y)
        if isinstance(block, Item) :
            block.touch(self, grid)

    @staticmethod
    def create(vector_x: float, vector_y: float, type_id: int):
        return Player(vector_x, vector_y, type_id, FieldOfView(0), "123")#uuid.uuid4())



import keyboard
from ..engine import Server

class Command:

    def apply(self, server: Server):
        server.player_stop_turning_left("123")
        server.player_stop_turning_right("123")
        server.player_stop_shooting("123")
        server.player_stop_moving_front("123")
        server.player_stop_moving_back("123")
        server.player_stop_moving_left("123")
        server.player_stop_moving_right("123")
        server.player_stop_interacting("123")
            
        if keyboard.is_pressed("left"):
            server.player_start_turning_left("123")
        if keyboard.is_pressed("right"):
            server.player_start_turning_right("123")
        if keyboard.is_pressed("up"):
            server.player_start_shooting("123")
        if keyboard.is_pressed("w"):
            server.player_start_moving_front("123")
        if keyboard.is_pressed("s"):
            server.player_start_moving_back("123")
        if keyboard.is_pressed("a"):
            server.player_start_moving_left("123")
        if keyboard.is_pressed("d"):
            server.player_start_moving_right("123")
        if keyboard.is_pressed(" "):
            server.player_start_interacting("123")



import os
import time
import win32console
import win32con
from ..engine import Server
from ..engine.config import FRAME_PER_SECONDS
from .command import Command
from .screen import Screen


class Game:
    def __init__(self, command: Command, screen: Screen, server: Server):
        self.__command = command
        self.__screen = screen
        self.__server = server
        
        self.__console = win32console.CreateConsoleScreenBuffer(DesiredAccess = win32con.GENERIC_READ | win32con.GENERIC_WRITE, ShareMode=0, SecurityAttributes=None, Flags=1)
        self.__console.SetConsoleActiveScreenBuffer()
        self.__server.load_map_file(
            # f'{os.path.dirname(__file__)}/../maps_pattern/map_test.txt')
            f'{os.path.dirname(__file__)}/../maps_pattern/map_1_level_1.txt')
    
        self.__frame_count = 0


    def start(self):
        self.__frame_count += 1
        loop_time = (1 / FRAME_PER_SECONDS)
        # self.__server.start_game()
        delta_time = loop_time
        self.__screen.draw_hud(self.__console)
        while 1:
            start_time = time.time()
            self.__command.apply(self.__server)
            self.__server.update(loop_time)
            state = self.__server.get_player_state("123")
            self.__screen.draw(self.__console, state)
            end_time = time.time()
            delta_time = end_time - start_time
            if delta_time < loop_time:
                time.sleep(loop_time - delta_time)
                delta_time = loop_time
            win32console.SetConsoleTitle(f'Pyfenstein3d - frame: {"%04.1f"%(1/delta_time)} - health: {"%03.0f"%state.health} - ammor: {"%03.0f"%state.ammo} - score: {state.score}')




from .item import Item

class Decoration(Item):
    def __init__(self, vector_x: float, vector_y: float, type_id: int, is_solid: bool):
        super().__init__(vector_x, vector_y, type_id, is_solid)



import math
from .person import Person
from .field_of_view import FieldOfView
from .vector2d import Vector2d

class Enemy(Person):
    def __init__(self, vector_x: float, vector_y: float, type_id: int, fov: FieldOfView):
        super().__init__(vector_x, vector_y, type_id, fov)
        self.__health = 100

    def _get_bounds(self, fov_ang: float):
        bounds = []
        pos = self._vector2d ** -fov_ang
        bounds.append(Vector2d(pos.x - 0.5, pos.y) ** fov_ang)
        bounds.append(Vector2d(pos.x + 0.5, pos.y) ** fov_ang)
        return bounds

    def get_state(self, player):
        ang = (player.fov_ang - (self.fov_ang + math.pi) + (math.pi * 2)) % (math.pi * 2)
        i = math.pi / 4
        for s in range(8):
            if ang < (s + 0.5) * i:
                return s
        return 0

    @staticmethod
    def create(vector_x: float, vector_y: float, type_id: int):
        if type_id == 131:
            return EnemyDog(vector_x, vector_y, FieldOfView(0))
        return EnemyGuard(vector_x, vector_y, FieldOfView(0))

class EnemyGuard(Enemy):
    def __init__(self, vector_x: float, vector_y: float, fov: FieldOfView):
        super().__init__(vector_x, vector_y, 130, fov)

class EnemyDog(Enemy):
    def __init__(self, vector_x: float, vector_y: float, fov: FieldOfView):
        super().__init__(vector_x, vector_y, 131, fov)



import math
from .vector2d import Vector2d
from .wall import Wall
from .item import Item
from .door import Door


class Ray():
    def __init__(self, angle: float):
        self.__dist_adjusted = 0
        self.__dist = 0
        self.__offset: 0
        self.__is_vertical: False
        self.__is_inverted: False
        self.__type_id: int = None
        self.__rel_ang = angle
        self.__vector2d_ang = Vector2d(1, 0)
        self.__vector2d = Vector2d(0, 0)
        self.__items = []
        self.__doors = []
        self.rot(angle)

    @property
    def ang(self):
        return self.__vector2d_ang.ang

    @property
    def dir_x(self):
        return self.__dir_x

    @property
    def dir_y(self):
        return self.__dir_y

    @property
    def delta_dist_x(self):
        return self.__delta_dist_x

    @property
    def delta_dist_y(self):
        return self.__delta_dist_y

    @property
    def type_id(self):
        return self.__type_id

    @property
    def offset(self):
        return self.__offset

    @property
    def is_vertical(self):
        return self.__is_vertical

    @property
    def is_inverted(self):
        return self.__is_inverted

    @property
    def dist_adjusted(self):
        return self.__dist_adjusted

    @property
    def dist(self):
        return self.__dist

    @property
    def collided_vector2d(self) -> Vector2d:
        return self.__vector2d

    @property
    def items(self):
        return tuple(self.__items)

    @property
    def doors(self):
        return tuple(self.__doors)

    def cast_wall(self, pos: Vector2d, grid):
        dist_x = self.get_dist_x(pos)
        dist_y = self.get_dist_y(pos)
        block_x = math.floor(pos.x)
        block_y = math.floor(pos.y)
        while 0 <= block_x < grid.max_x or 0 <= block_y < grid.max_y:
            if dist_y < dist_x:
                block_y += self.dir_y
                item = grid.get_block(block_x, block_y)
                if isinstance(item, Wall):
                    self.__vector2d = Vector2d(
                        math.cos(self.ang) * dist_y,
                        math.sin(self.ang) * dist_y
                    ) + pos
                    self.__dist_adjusted = abs(
                        math.cos(self.__rel_ang) * dist_y)
                    self.__dist = abs(dist_y)
                    self.__is_inverted = pos.y > item.y
                    self.__offset = self.__vector2d.x % 1
                    self.__is_vertical = False
                    if self.__is_inverted and isinstance(grid.get_block(block_x, block_y + 1), Door):
                        self.__type_id = 50
                    elif not self.__is_inverted and isinstance(grid.get_block(block_x, block_y - 1), Door):
                        self.__type_id = 50
                    else:
                        self.__type_id = item.type_id

                    return
                dist_y += self.delta_dist_y
            else:
                block_x += self.dir_x
                item = grid.get_block(block_x, block_y)
                if isinstance(item, Wall):
                    self.__vector2d = Vector2d(
                        math.cos(self.ang) * dist_x,
                        math.sin(self.ang) * dist_x
                    ) + pos
                    self.__dist_adjusted = abs(
                        math.cos(self.__rel_ang) * dist_x)
                    self.__dist = abs(dist_x)
                    self.__is_inverted = pos.x > item.x
                    self.__offset = self.__vector2d.y % 1
                    self.__is_vertical = True
                    if self.__is_inverted and isinstance(grid.get_block(block_x + 1, block_y), Door):
                        self.__type_id = 50
                    elif not self.__is_inverted and isinstance(grid.get_block(block_x - 1, block_y), Door):
                        self.__type_id = 50
                    else:
                        self.__type_id = item.type_id
                    return
                dist_x += self.delta_dist_x

    def cast_doors(self, pos: Vector2d, doors: []):
        self.__doors = []
        for door in doors:
            if door.is_vertical:
                is_inverted = math.pi * 0.5 <= self.ang <= math.pi * 1.5
                if not is_inverted and  door.x > self.__vector2d.x:
                    continue
                if is_inverted and door.x < self.__vector2d.x:
                    continue
                x = door.x - pos.x;
                y = math.tan(self.ang) * x;
                v = Vector2d(x, y);
                pos_y = pos.y + y;
                if pos_y < door.y or pos_y > door.y + 1:
                    continue
                dist_adjusted = math.sin(math.pi / 2 - self.__rel_ang) * v.mag;
                offset = (door.y - pos_y) % 1 #- Math.floor(door.y - pos_y);
                self.__doors.append(RayDoor(dist_adjusted, offset, door.is_vertical))
                if v.mag < self.__dist:
                    self.__dist = v.mag
                    self.__vector2d = v + pos
                    self.__dist_adjusted = dist_adjusted
            else:
                is_inverted = math.pi <= self.ang <= math.pi * 2
                if not is_inverted and door.y > self.__vector2d.y:
                    continue
                if is_inverted and door.y < self.__vector2d.y:
                    continue
                y = door.y - pos.y
                x = y / math.tan(self.ang)
                v = Vector2d(x, y)
                pos_x = pos.x + x
                if pos_x < door.x or pos_x > door.x + 1:
                    continue
                dist_adjusted = math.sin(math.pi / 2 - self.__rel_ang) * v.mag
                offset = (door.x - pos_x) % 1
                self.__doors.append(RayDoor(dist_adjusted, offset, door.is_vertical))
                if v.mag < self.__dist:
                    self.__dist = v.mag
                    self.__vector2d = v + pos
                    self.__dist_adjusted = dist_adjusted

    def cast_items(self, player, items: []):
        self.__items = []
        pos = Vector2d(player.x, player.y)
        if self.type_id is not None:
            for i in items:
                item: Item = i
                item_pos = (Vector2d(item.x, item.y) - pos)
                ray_pos = (self.__vector2d - pos)
                if item_pos.mag < ray_pos.mag:
                    ray_pos = ray_pos ** -self.ang
                    item_pos = item_pos ** -self.ang
                    if ray_pos.x < item_pos.x:
                        continue
                    if item_pos.y > 0.5 or item_pos.y < -0.5:
                        continue
                    dist = math.sin(math.pi / 2 - self.__rel_ang) * item_pos.x
                    self.__items.append(RayItem(item.type_id, dist, item_pos.y, item.get_state(player)))
                    
                

    def rot(self, rad):
        self.__vector2d_ang = self.__vector2d_ang ** rad

        if self.ang > math.pi * 1.5 or self.ang < math.pi * 0.5:
            self.__dir_x = 1
        elif self.ang > math.pi * 0.5 and self.ang < math.pi * 1.5:
            self.__dir_x = -1
        else:
            self.__dir_x = 0

        if self.ang > 0 and self.ang < math.pi:
            self.__dir_y = 1
        elif self.ang > math.pi and self.ang < math.pi * 2:
            self.__dir_y = -1
        else:
            self.__dir_y = 0

        self.__delta_dist_x = abs(
            1 / math.cos(self.ang)) if math.cos(self.ang) != 0 else math.inf
        self.__delta_dist_y = abs(
            1 / math.sin(self.ang)) if math.sin(self.ang) != 0 else math.inf

    def get_dist_x(self, pos: Vector2d):
        if self.dir_x == 1:
            return abs((1 - pos.x % 1) / math.cos(self.ang))
        if self.dir_x == -1:
            return abs((pos.x % 1) / math.cos(self.ang))
        return math.inf

    def get_dist_y(self, pos: Vector2d):
        if self.dir_y == 1:
            return abs((1 - pos.y % 1) / math.sin(self.ang))
        if self.dir_y == -1:
            return abs((pos.y % 1) / math.sin(self.ang))
        return math.inf


class RayItem():
    def __init__(self, type_id, dist, offset, state):
        self.__type_id = type_id
        self.__dist = dist
        self.__offset = offset
        self.__state = state

    @property
    def type_id(self):
        return self.__type_id

    @property
    def dist(self):
        return self.__dist

    @property
    def offset(self):
        return self.__offset

    @property
    def state(self):
        return self.__state

class RayDoor():
    def __init__(self, dist, offset, is_vertical):
        self.__type_id = 49
        self.__dist = dist
        self.__offset = offset
        self.__is_vertical = is_vertical

    @property
    def type_id(self):
        return self.__type_id

    @property
    def dist(self):
        return self.__dist

    @property
    def offset(self):
        return self.__offset

    @property
    def is_vertical(self):
        return self.__is_vertical

    @property
    def state(self):
        return 1 if self.is_vertical else 0



from .vector2d import Vector2d
from .item import Item
from .wall import Wall
from .decoration import Decoration
from .person import Person
from .field_of_view import FieldOfView
from .map2d import Map2d
from . import config
from .item_grid import ItemGrid
from .ray import Ray
from .player import Player
from .server import Server
from .animation import Animation
from .weapon import Weapon
from .weapon import WeaponPistol
from .door import Door



from numpy import interp

class Animation:
    def __init__(self, time: float, on_animate=None, on_animate_end=None):
        self.__total_time = time
        self.__on_animate = on_animate
        self.__on_animate_end = on_animate_end
        self.__time = 0
        self.__factor = 0
        self.__is_animating = False

    @property
    def total_time(self):
        return self.__total_time;

    @property
    def time(self):
        return self.__time;

    @property
    def is_animating(self):
        return self.__is_animating;

    @property
    def factor(self):
        return self.__factor;

    def start(self, at: int = 0):
        self.__time = at
        self.__is_animating = True

    def update(self, delta_time: float):
        if self.__is_animating:
            self.__time += delta_time
            self.__factor =  self.__time / self.__total_time
            self.__factor = self.__factor if self.__factor <= 1 else 1;
            if self.__on_animate is not None:
                self.__on_animate(self.__factor)
            if self.__factor >= 1:
                self.__is_animating = False
                if self.__on_animate_end is not None:
                    self.__on_animate_end()
        else:
            self.__factor = 0



from .animation import Animation

class Weapon():
    def __init__(self, type_id:int, shoot_interval:int):
        self.__type_id = type_id
        self.__shoot_interval = shoot_interval;
        self.__shoot_animation = Animation(0.5)
        self.is_shooting = False

    @property
    def type_id(self):
        return self.__type_id

    @property
    def shoot_interval(self):
        return self.__shoot_interval

    @property
    def shoot_animation(self):
        return self.__shoot_animation

    def update(self, delta_time:float, person, grid):
        if self.is_shooting and (not self.__shoot_animation.is_animating or self.__shoot_animation.time > self.__shoot_interval):
            self.__shoot_animation.start()
            person.shoot()
        self.__shoot_animation.update(delta_time)


class WeaponKnife(Weapon):
    def __init__(self):
        super().__init__(120, 0.5)

class WeaponPistol(Weapon):
    def __init__(self):
        super().__init__(121, 0.5)

class WeaponSubmachine(Weapon):
    def __init__(self):
        super().__init__(122, 0.3)

class WeaponMinigun(Weapon):
    def __init__(self):
        super().__init__(123, 0.1)



from .item import Item
from .wall import Wall
from .decoration import Decoration
from .item_grid import ItemGrid
from .player import Player
from .person import Person
from .enemy import Enemy
from .door import Door
from .item_weapon import ItemWeapon
from .item_ammo import ItemAmmo
from .item_score import ItemScore


class Map2d():
    def __init__(self, items: []):
        self.__grid = ItemGrid(items)
        self.__players = {item.player_id: item for item in items if isinstance(item, Player)}
        self.__persons = [item for item in items if isinstance(item, Person)]
        self.__doors = [item for item in items if isinstance(item, Door)]
        self.__enemies = [item for item in items if isinstance(item, Enemy)]

    @property
    def grid(self) -> Item:
        return self.__grid

    def update(self, delta_time: float):
        for player_id in self.__players:
            player = self.__players[player_id]
            player.update(delta_time, self.__grid)
            player.adjust_collision(self.__grid)
            player.cast(self.__grid)
        for door in self.__doors:
            door.update(delta_time, self.__persons)


    def get_player(self, player_id: str):
        return self.__players[player_id]

    @staticmethod
    def create_with_pattern(pattern: str):
        items = []
        lines = pattern.splitlines()
        list_type_ids = [[s[n:n+2]
                          for n in range(0, len(s), 2)] for s in lines]
        for block_y, type_ids in enumerate(list_type_ids):
            for block_x, type_id_hex in enumerate(type_ids):
                if type_id_hex.strip() == "":
                    continue
                type_id = int(type_id_hex, 16)
                if type_id in type_ids_walls_solid:
                    items.append(Wall(block_x, block_y, type_id))
                elif type_id in type_ids_weapon:
                    items.append(ItemWeapon(block_x + 0.5, block_y + 0.5, type_id, False))
                elif type_id in type_ids_ammo:
                    items.append(ItemAmmo(block_x + 0.5, block_y + 0.5, type_id, False))
                elif type_id in type_ids_score:
                    items.append(ItemScore(block_x + 0.5, block_y + 0.5, type_id, False))
                elif type_id in type_ids_decorations_non_solid:
                    items.append(Decoration(block_x + 0.5, block_y + 0.5, type_id, False))
                elif type_id in type_ids_decorations_solid:
                    items.append(Decoration(block_x + 0.5, block_y + 0.5, type_id, True))
                elif type_id in type_ids_doors_horizontal:
                    items.append(Door(block_x, block_y + 0.5, type_id, False))
                elif type_id in type_ids_doors_vertical:
                    items.append(Door(block_x + 0.5, block_y, type_id, True))
                elif type_id in type_ids_enemies:
                    items.append(Enemy.create(block_x + 0.5, block_y + 0.5, type_id))
                elif type_id in type_ids_player:
                    items.append(Player.create(block_x + 0.5, block_y + 0.5, type_id))
        return Map2d(items)



type_ids_decorations_solid = [57, 60, 61, 63, 65, 66, 68,
                         69, 70, 71, 74, 75, 76, 80, 93, 94, 95, 97]
# type_ids_decorations_non_solid = [58, 62, 67,
#                              72, 73, 77, 81, 92, 96, 99, 100, 101, 102]

type_ids_decorations_non_solid = [i for i in range(56,120) if i not in type_ids_decorations_solid]
type_ids_walls_solid = list(range(48)) + [53, 54]
type_ids_score = [ 87, 88, 89 ]
type_ids_weapon = [ 85, 86 ]
type_ids_ammo = [ 84 ]
type_ids_doors_horizontal = [50]
type_ids_doors_vertical = [49]
type_ids_player = [255]
type_ids_enemies = [130, 131]



import setuptools


with open('README.md') as f:
    readme = f.read()

with open("requirements.txt") as f:
    install_requires = f.read().splitlines()

setuptools.setup(name='pyfenstein3d',
                 version='1.4.2',
                 description='Projeto realizado em python com intuido educacional de tentar reproduzir o jogo Wolfenstein 3d no prompt de comando.',
                 long_description=readme,
                 long_description_content_type='text/markdown',
                 license='MIT License',
                 author='Allison GrrriiiM',
                 author_email='allison.f.oliveira@hotmail.com',
                 url='https://github.com/GrrriiiM/pyfenstein3d',
                 packages=setuptools.find_packages(exclude=["tests"]),
                 package_data={'pyfenstein3d': ['imgs/*.*', 'maps_pattern/*.*']},
                 install_requires=install_requires,
                 include_package_data=True,
                 )



if __name__ == "__main__":
    import os
    from .engine import Server
    from .game import Command
    from .game import Screen
    from .game import Game
    from .game import Image
    os.system("cls")
    print("CONFIGURACOES")
    print("- FONTE: CONSOLAS 10px")
    print("- LARGURA: 200")
    print("- ALTURA: 62")
    print()
    print("COMANDOS")
    print("- Andar para frente: W")
    print("- Andar para trás: S")
    print("- Andar para esquerda: A")
    print("- Andar para direita: D")
    print("- Virar para esquerda: Seta esquerda")
    print("- Virar para direita: Seta direta")
    print("- Atirar: Seta cima")
    print("- Abrir porta: Espaço")
    print()
    print("Pressione qualquer tecla para começar")
    input()
    command = Command()
    image = Image()
    screen = Screen(image)
    server = Server()
    game = Game(command, screen, server)
    game.start()



from .item import Item

class ItemScore(Item):
    def touch(self, player, grid):
        if self.type_id == 87:
            player.add_score(100)
        elif self.type_id == 88:
            player.add_score(400)
        elif self.type_id == 89:
            player.add_score(800)
        grid.remove_block(self.x, self.y)



from .vector2d import Vector2d
from .block import Block

class Item(Block):
    def __init__(self, vector_x: float, vector_y: float, type_id: int, is_solid: bool=False):
        super().__init__(vector_x, vector_y, type_id, is_solid, False)

    def _get_bounds(self, fov_ang: float):
        bounds = []
        pos = self._vector2d ** -fov_ang
        bounds.append(Vector2d(pos.x - 0.5, pos.y) ** fov_ang)
        bounds.append(Vector2d(pos.x + 0.5, pos.y) ** fov_ang)
        return bounds

    def touch(self, player, grid):
        pass
